from typing import Dict, Tuple, Union


KEYS = tuple(
    "# < ^ S- K- F- P- T- N- R- *- E- A- -I -U "
    "-R -N -P -S -K -F -T -e -_ -*".split()
)

IMPLICIT_HYPHEN_KEYS = ("E-", "A-", "-I", "-U")

SUFFIX_KEYS = ("-e", )

NUMBER_KEY = None

NUMBERS = {}

FERAL_NUMBER_KEY = False

UNDO_STROKE_STENO = ""

ORTHOGRAPHY_RULES = []

ORTHOGRAPHY_RULES_ALIASES = {}

ORTHOGRAPHY_WORDLIST = None

DICTIONARIES_ROOT = "asset:plover_delta_de:dictionaries"

DEFAULT_DICTIONARIES = tuple()


def build_map(
    system_keys: str, 
    machine_keys: str
) -> Dict[str, Union[str, Tuple[str, ...]]]:
    system_keys_list = system_keys.split()
    machine_map = {s: [] for s in system_keys_list}
    for m, s in zip(machine_keys.split(), system_keys_list):
        machine_map[s].append(m)
    
    return dict(
        (m, l[0]) if len(l) == 1 else (m, tuple(l))
        for m, l in machine_map.items()
    )

KEYMAP_LAYOUTS = {
    "Gemini PR": ((
        "S1- T- P- H- *1  *3 -F -P -L -T -D "
        "S2- K- W- R- *2  *4 -R -B -G -S -Z "
                  "A- O-  -E -U"
    ), (
        "#1", "#2", "#3", "#4", "#5", "#6", "#7", 
        "#8", "#9", "#A", "#B", "#C"
    ), (
        "Fn", "pwr", "res1", "res2"
    ), None),

    "Keyboard": ((
        "q  w  e  r  t   y  u  i  o  p  [ "
        "a  s  d  f  g   h  j  k  l  ;  ' "
                 "v  b   n  m"
    ), (
        "1", "2", "3", "4", "5", "6", "7", "8", 
        "9", "0", "-", "=", 
    ), (
        ",", ".", "/", "]", "\\", "z", 
        "x"
    ), "space"),

    "Treal": ((
        "X1- S1- T- P- H-  *1 -F -P -L -T -D "
        "X2- S2- K- W- R-  *2 -R -B -G -S -Z "
                   "A- O-  -E -U"
    ), (
        "#1", "#2", "#3", "#4", "#5", "#6", "#7", 
        "#8", "#9", "#A", "#B"
    ), (
        "X3",
    ), None),
}

LAYOUT = (
    "T- R- S- *- -_   ^ -* -N -S -F -e "
    "N- P- K- F- <    ^ -R -P -K -T -e "
             "E- A-  -I -U"
)

KEYMAPS = {
    map_name: {
        **build_map(LAYOUT, map_layout),
        "#": num_keys,
        "no-op": noop_keys,
        **({"arpeggiate": arp_key} if arp_key else {})
    }
    for map_name, (map_layout, num_keys, noop_keys, arp_key)
    in KEYMAP_LAYOUTS.items()
}



import json
from typing import Tuple

from plover.steno_dictionary import StenoDictionary
from plover_stroke import BaseStroke

from plover_delta_de.system import KEYS, IMPLICIT_HYPHEN_KEYS


class Stroke(BaseStroke):
    @classmethod
    def setup(cls, keys: tuple, implicit_hyphen_keys: tuple):
        cls.keys = keys
        cls.ihks = implicit_hyphen_keys

        medial_pos_list = [list(cls.keys).index(k) for k in cls.ihks]
        cls.medial_pos = min(medial_pos_list)
        cls.final_pos = max(medial_pos_list)

        cls.init_key_order = {
            k.replace("-", ""): n
            for n, k in enumerate(list(cls.keys)[:cls.medial_pos])
        }
        cls.post_key_order = {
            k.replace("-", ""): n + cls.medial_pos
            for n, k in enumerate(list(cls.keys)[cls.medial_pos:])
        }
        cls.medial_keys = set(k.replace("-", "") for k in cls.ihks)

        super().setup(keys, implicit_hyphen_keys)

    @classmethod
    def from_stroke(cls, stroke: str) -> str:
        as_bin = 0
        order = cls.init_key_order
        in_init = True
        in_final = False
        first_e = False
        for char in stroke:
            if in_init and (char == "-" or char in cls.medial_keys):
                order = cls.post_key_order
                in_init = False
            
            if not in_init and not in_final:
                if char not in cls.medial_keys:
                    in_final = True
                elif char == "E" and first_e:
                    char = "e"
            elif in_final and char in cls.medial_keys:
                if char == "E":
                    char = "e"
                else:
                    raise ValueError(
                        "Invalid stroke: Medial in wrong position"
                    )
            
            if char == "E":
                first_e = True

            if char != "-":
                to_add = 1 << order[char]
                if to_add & as_bin:
                    raise ValueError("Invalid stroke: Duplicate key")

                as_bin += to_add
        
        return Stroke(as_bin)


Stroke.setup(KEYS, IMPLICIT_HYPHEN_KEYS)


class DeltaDictionary(StenoDictionary):
    readonly = False

    def __init__(self) -> None:
        super().__init__()
        self._reorder_map = {}
    
    def _load(self, filename: str) -> None:
        with open(filename, "r", encoding="utf-8") as fp:
            json_data = json.load(fp)
        
        for unordered, output in json_data.items():
            ordered = tuple(
                str(Stroke.from_stroke(stroke))
                for stroke in
                unordered.split("/")
            )

            self._reorder_map[ordered] = unordered
            self[ordered] = output
    
    def _save(self, filename: str) -> None:
        mappings = []
        for strokes, translation in self.items():
            ordered = "/".join(strokes)
            mappings.append((
                self._reorder_map.get(
                    ordered, 
                    ordered.replace("e", "E")
                ),
                translation
            ))
        
        mappings.sort()

        with open(filename, "w", encoding="utf-8", newline="\n") as fp:
            json.dump(
                dict(mappings),
                fp,
                ensure_ascii=False,
                indent=0,
                separators=(",", ": ")
            )

            fp.write("\n")
    
    def __contains__(self, key: Tuple[str]) -> bool:
        return False
    
    def get(self, key: Tuple[str], fallback=None) -> str:
        if len(key) > self._longest_key:
            return fallback
        
        if key in self._dict:
            return self[key]

        capitalized = "^" in key[0]
        attach = "<" in key[0]
        if not (capitalized or attach):
            return fallback
        
        new_key = (
            key[0].replace("^", "").replace("<", ""),
            *key[1:]
        )

        if new_key not in self._dict:
            return fallback
        
        return (
            "{^}" * attach +
            "{-|}" * capitalized +
            self[new_key]
        )


def split_entry(line: str) -> Tuple[str, str]:
    word = ""
    last = ""
    for i, c in enumerate(line):
        if c == "," and last != "\\":
            return (
                word.replace("\,", ",").strip(), 
                line[i+1:].strip()
            )

        word += c
        last = c
    
    # Fallback
    word, outline = line.strip().split(",", 1)
    return word.strip(), outline.strip()


class DeltaWordDictionary(DeltaDictionary):
    def _load(self, filename: str) -> None:
        with open(filename, "r", encoding="utf-8") as fp:
            for line in fp.readlines():
                if "," not in line:
                    continue

                word, unordered = split_entry(line)
                ordered = tuple(
                    str(Stroke.from_stroke(stroke))
                    for stroke in
                    unordered.split(".")
                )

                self._reorder_map[ordered] = unordered
                self[ordered] = word.strip()
    
    def _save(self, filename: str) -> None:
        mappings = []
        for strokes, translation in self.items():
            ordered = "/".join(strokes)
            mappings.append((
                translation.replace(",", "\,"),
                self._reorder_map.get(
                    ordered, 
                    ordered.replace("/", ".").replace("e", "E")
                )
            ))
        
        mappings.sort()

        with open(filename, "w", encoding="utf-8", newline="\n") as fp:
            for (translation, outline) in mappings:
                fp.write(f"{translation}, {outline}\n")
